#include <unistd.h>
#include <stdio.h>
#include <math.h>

int jacobi(int N, double *T[], double *dT[], 
           int maxits, int printfreq, double tolerance,
           int* iterations, double *finalerror)
{
//----------------------------------------------------------------------------------
// Do a Jacobi iteration for a pure diffusion process on a (N+2) x (N+2) mesh of
// nodes. The extremes at left/right/top/bottom of the plate are indexed by 0 and
// N+1, and correspond to boundary conditions which are already set (and so should
// not be updated at all).
//
// Every printfreq iterations, the error norm is printed out to a file called
// "history". That should be suitable for plotting using Matlab or damned near
// anything else at hand. The Matlab script plothistory.m will read it in and do 
// the plotting if you like. If printfreq <= 0, the history file is not opened or 
// written to. That should be the setting for making performance analysis tests, to
// avoid perturbations from I/O in the iteration.
//
// At the end, the error norm and number of iterations is fed back to the caller
// via the last two pointer arguments.
//----------------
// Randall Bramley
// Department of Computer Science
// Indiana University, Bloomington
//----------------
// Last Modified: Wed 24 Jan 2018, 12:57 PM
//----------------------------------------------------------------------------------

    int talkalot = 0;                // Verbosity control
    int i, j;                        // loop indices
    double errornorm = __DBL_MAX__ ; // initialize error norm to humongous
    int iteration = 0;               // number of iterations so far
    FILE *history;

    if (printfreq > 0){
        history = fopen("history", "a");
        if (history == NULL) return -5;
    }

    while (iteration <= maxits && errornorm > tolerance) {

        // Get the update for each point
        for (i = 1; i < N+1; ++i) 
            for (j = 1; j < N+1; ++j)
                dT[i][j] = (     T[i-1][j]        // neighbor to left
                              +  T[i+1][j]        // neighbor to right
                              +  T[i]  [j-1]      // neighbor below
                              +  T[i]  [j+1])/4.0 // neighbor above
                              -  T[i][j];

        // Apply the update to interior of T (non-boundary nodes)
        for (i = 1; i < N+1; ++i) 
            for (j = 1; j < N+1; ++j)
                T[i][j] += dT[i][j];
        
        // Compute the error norm
        errornorm = 0.0;
        for (i = 1; i < N+1; ++i) 
            for (j = 1; j < N+1; ++j)
                errornorm += fabs(dT[i][j]);
        errornorm = errornorm/(N*N);

        // That's one small step for an iteration, a giant leap for ants
        iteration += 1;
        if ((printfreq > 0) && iteration%printfreq == 1) 
            fprintf(history, "%d  %.17e\n", iteration, errornorm);
    }

    // Make sure the last iterate is printed, even if it is not a multiple of
    // printfreq.
    if ((printfreq > 0) && iteration%printfreq != 1)
            fprintf(history, "%d  %.17e\n", iteration, errornorm);

    *iterations = iteration;
    if (talkalot) printf("jacobi: iterations %d\n", *iterations);
    *finalerror = errornorm;
    if (talkalot )printf("jacobi: finalerror %e\n", *finalerror);

    if (printfreq > 0) fclose(history);

    return 0;

}
