#include <unistd.h>
#include <stdio.h>
#include <stddef.h>  // for size_t
#include <stdlib.h>  // free
#include <string.h>
#include <time.h>
#include <omp.h>

// Headers. I'm too lazy to use the .h convention.
double elapsedtime();
int getparms(int *N, int *maxits, double *tolerance, int *printfreq, 
        int *showplate, int*method);
int BC(int N, double *T[]);
int initialize(int N, double *T[], double *dT[]);
int printplate(int N, double *T[]);
int jacobi(int N, double *T[], double *dT[], 
           int maxits, int printfreq, double tolerance,
           int* iterations, double *finalerror);

int main(int argc, char **argv) {
//-----------------------------------------------------------------------------------
// Solve heat transfer problem in a discretized plate, with mesh lines in both x and
// y directions numbered from 0 to N+1. The two extremes (0 and N+1) correspond to
// boundary conditions, so those are initialized but not updated during the
// diffusion process. In addition to the problem size and control parameters in the
// file "parameters", this code can also output the final values of the plate
// temperatures to a file named "plate". The Matlab script plotplate.m will then
// read it in and, surprise, plot it. However, for performance analysis keep the 
// call to printplate commented out.
//
// The overall process is
//
//      read in parameters for mesh size and iteration control
//      set the boundary nodes to values given by BC.c [= boundary conditions]
//      initialize the interior nodes to all zeros
//      time the jacobi iteration in jacobi.c
//      print out some basic information to std out
//      optionally print out the final values
//
// Although most of the functions return an error flag, after debugging I've omitted
// the testing of those on return to reduce clutter. Add them back in if things go
// wrong. The most likely thing is to fail to provide the input file "parameters",
// so that test remains here.
//
//-----------------------
// Input/output files: 
//-----------------------
//
//  parameters: contains 6 control variables as input
//
//  history: has iteration number and error measure for chosen iterates; see
//      the variable printfreq. Can be plotted using the Matlab script plothistory.m.
//      This main() function will write out some preliminary and final lines to that
//      file, but preceded by the Matlab comment indicator "%" so they do not affect
//      using the history file to plot convergence.
//
//  plate: if showplate is true, will contain the final iterate. Can be plotted
//      using the Matlab script showplate.m
//
//  results: a file with condensed performance information. use the matlab script
//      showresults.m to plot them, but beware it is only valid for runs where all
//      of N, tolerance, and maxits are the same.
//
//--------------------
// Input parameters:
//--------------------
//  N = number of unknowns in both x and y direction. The plate size is actually
//      N+2 x N+2, when boundary nodes are also counted
//
//  maxits = maximum allowed number of iterations
//
//  tolerance = stopping test tolerance on the error measure, which is the 2-norm 
//      of the update size divided by N^2.
//
//  printfreq = how often to write out the error measure. Set to a negative 
//      number to avoid having it printed out at all.
//
//  showplate = Boolean, whether or not to output the final iterate to a file
//
// The variable "method" is intended to take on some value in the set {1..4}:
//  1. Use Jacobi method, no OpenMP
//  2. Use Jacobi method, with OpenMP
//  3. Use Gauss-Seidel method, no OpenMP
//  4. Use Gauss-Seidel method, with OpenMP
//
//-----------
// C notes: 
//-----------
// To make sure that the data is located contiguously, first the (N+2)^2 doubles
// are allocated to mT in a single call to calloc(). Then T is defined as a vector
// of pointers to 1D arrays, with T[i] a pointer to row i in mT. Mutatis mutandi for
// the dT array. This allows double indexing in the called functions, at the cost of
// two arrays of pointers of length N+2 each.
//
//----------------
// Randall Bramley
// Department of Computer Science
// Indiana University, Bloomington
//----------------
// Last Modified: Thu 25 Jan 2018, 05:32 PM
//-----------------------------------------------------------------------------------

    int writesummary = 1;   // Verbosity control
    int i, j, k, errflag, nthreads; 

	// Problem specs
    int n, N, maxits, printfreq, showplate, method;
    double tolerance;
    double *mT;
    double *mdT;
    double **T;
    double **dT;
    // Final values from the Jacobi iterations
    int *iterations    = malloc(sizeof(int));
    double *finalerror = malloc(sizeof(double));

    // Variable for file handling
    FILE *parameters, *history, *results;

    // Timing variables
    double time1, time2, t, gflops;

    //=============================================================================
    //================ Executable statements start ================================
    //=============================================================================

    // Read in the mesh size and iteration control parameters
    errflag = getparms(&N, &maxits, &tolerance, &printfreq, &showplate, &method);
    if (errflag) {
        printf("Bum value from getparms; errflag = %d\n", errflag);
        return -1;
    }

    // Allocate array storage as a single blob for both T and dT
    n = (N+2)*(N+2);    // n = total number of nodes
    mT  = calloc(n, sizeof(double));
    mdT = calloc(n, sizeof(double));
    if (mT == NULL || mdT == NULL) {
        printf("Could not allocate mT and mdT; check sizes\n");
        printf("N = %d and n = %d\n", N, n);
        return -2;
    }

    // Set up T and dT to allow double indexing
    T  = (double **) calloc(N+2, sizeof(double *));
    dT = (double **) calloc(N+2, sizeof(double *));
    if (T == NULL || dT == NULL) {
        printf("Could not allocate T and dT\n");
        printf("N = %d and n = %d\n", N, n);
        return -3;
    }
    // Point T[i] to i-th row in the mT blob
    for (i = 0; i < N+2;  ++i) {
        T[i]  = mT  + i*(N+2);
        dT[i] = mdT + i*(N+2);
    }

    // Have to start a parallel region to get number of threads started by default
#pragma omp parallel
    nthreads = omp_get_num_threads();

    // Tell me a story about the inputs, grandpa.
    if (writesummary){

        if (printfreq > 0){
            history = fopen("history", "w");
            if (history == NULL) return -5;
            fprintf(history, "\n%% Max allowed number of iterations = %d \n", maxits);
            fprintf(history, "%% Stopping test tolerance = %e \n", tolerance);
            fprintf(history, "%% N = %d and number of unknowns = %d\n", N, n);
            fprintf(history, "%% Total number of nodes = %d\n", n);
        }

        printf("============================================================\n");
        printf("Max allowed number of iterations = %d \n", maxits);
        printf("Stopping test tolerance = %e \n", tolerance);
        printf("N = %d and number of unknowns = %d\n", N, n);
        printf("Total number of nodes = %d\n", n);


        if (showplate){
            printf("Solution will be written out to a file at end\n");}
        else{
            printf("Solution will not be written out at end\n");
        }
        if (printfreq > 0){
            printf("printfreq = %d \n", printfreq);
        }
        else {
            printf("Convergence history will not be saved\n");
        }
        printf("Number of OMP threads = %d \n", nthreads);
        if (printfreq > 0){
            fprintf(history, "%% Number of OMP threads = %d \n\n", nthreads);
            fclose(history);
        }
    }

    // Set boundary conditions
    BC(N, T);

    // Set initial conditions on interior of the domain. Using calloc
    // actually zeros out things, but this will make sure and can be
    // repeated if more than one solver method is tested.
    initialize(N, T, dT);

    time1 = elapsedtime();
    if (method == 1)
        jacobi(N, T, dT, maxits, printfreq, tolerance,
            iterations, finalerror);
    time2 = elapsedtime() - time1;

    // Process takes about 8N^2 flops per iteration, giving the
    // following for the computational rate in Gflops per sec
    gflops = *iterations*8.0e-9*(double)N*(double)N/time2;

    if (writesummary){
        if (printfreq > 0){
            history = fopen("history", "a");
            if (history == NULL) return -5;
            fprintf(history,"\n%% Time taken            = %.17e secs\n", time2);
            fprintf(history,"%% Gflops/sec rate       = %.17e \n", gflops);
        }
        printf("Number of iterations  = %d\n", *iterations);
        printf("Final error norm      = %.17e\n", *finalerror);
        printf("Time taken            = %.17e secs\n", time2);
        printf("Gflops/sec rate       = %.17e \n", gflops);
        printf("============================================================\n");

    }

    results = fopen("results", "a");
    if (results == NULL) return -5;
    fprintf(results, "%4d    %2d     %d   %.9e   %.9e   %.9e   %1d\n", 
            N, nthreads, *iterations, *finalerror, time2, gflops, method);
    fclose(results);

    // Print out values in the plate. Don't do this for large N, and
    // don't do it for performance tests.
    if (showplate) printplate(N, T);

    free( (void *) T);
    free( (void *) dT);
    free( (void *) mT);
    free( (void *) mdT);

    return 0;
}
